@@ -1,4 +1,3 @@
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -19,14 +18,12 @@
 
 
 #include "dvb_scan.h"
+#include "dvb_debug.h"
 
 
 /* ----------------------------------------------------------------------- */
 /* ----------------------------------------------------------------------- */
 /* maintain current state for these ... */
-//char *dvb_src   = NULL;
-//char *dvb_lnb   = NULL;
-//char *dvb_sat   = NULL;
 extern int  dvb_inv;
 
 /* ----------------------------------------------------------------------- */
@@ -123,8 +120,8 @@ typedef void (*dvbmon_notify)(struct psi_info *info, int event,
 #define DVBMON_EVENT_UPDATE_PR    3
 #define DVBMON_EVENT_DESTROY     99
 
-struct dvbmon* dvbmon_init(struct dvb_state *dvb, int verbose,
-			   int o_nit, int o_sdt, int pmts);
+//struct dvbmon* dvbmon_init(struct dvb_state *dvb, int verbose,
+//			   int o_nit, int o_sdt, int pmts);
 void dvbmon_refresh(struct dvbmon* dm);
 void dvbmon_add_callback(struct dvbmon* dm, dvbmon_notify func, void *data);
 void dvbmon_del_callback(struct dvbmon* dm, dvbmon_notify func, void *data);
@@ -144,75 +141,10 @@ static void table_close(struct dvbmon *dm, struct table *tab);
 static void table_del(struct dvbmon *dm, int pid, int sec);
 static void table_next(struct dvbmon *dm);
 
-
 /* ----------------------------------------------------------------------- */
 // DEBUG
 /* ----------------------------------------------------------------------- */
 
-/* ----------------------------------------------------------------------- */
-static void print_stream(struct psi_stream *stream)
-{
-
-	//    	int                  tsid;
-	//
-	//        /* network */
-	//        int                  netid;
-	//        char                 net[PSI_STR_MAX];
-	//
-	//        int                  frequency;
-	//        int                  symbol_rate;
-	//        char                 *bandwidth;
-	//        char                 *constellation;
-	//        char                 *hierarchy;
-	//        char                 *code_rate_hp;
-	//        char                 *code_rate_lp;
-	//        char                 *fec_inner;
-	//        char                 *guard;
-	//        char                 *transmission;
-	//        char                 *polarization;
-	//
-	//        /* status info */
-	//        int                  updated;
-	fprintf(stderr, "TSID %d NETID %d : network %s : freq %d : sr %d : BW %s : Const %s : Hier %s : Code rate hp %s lp %s : FEC %s : Guard %s : Tx %s : Pol %s (up %d, tuned %d)\n",
-			stream->tsid, stream->netid, stream->net, stream->frequency, stream->symbol_rate,
-			stream->bandwidth, stream->constellation, stream->hierarchy, stream->code_rate_hp, stream->code_rate_lp,
-			stream->fec_inner, stream->guard, stream->transmission, stream->polarization,
-			stream->updated, stream->tuned
-	) ;
-}
-
-/* ----------------------------------------------------------------------- */
-static void print_program(struct psi_program *program)
-{
-
-	//        int                  tsid;
-	//         int                  pnr;
-	//         int                  version;
-	//         int                  running;
-	//         int                  ca;
-	//
-	//         /* program data */
-	//         int                  type;
-	//         int                  p_pid;             // program
-	//         int                  v_pid;             // video
-	//         int                  a_pid;             // audio
-	//         int                  t_pid;             // teletext
-	//         char                 audio[PSI_STR_MAX];
-	//         char                 net[PSI_STR_MAX];
-	//         char                 name[PSI_STR_MAX];
-	//
-	//         /* status info */
-	//         int                  updated;
-	//         int                  seen;
-
-	fprintf(stderr, "TSID %d PNR %d : name %s : network %s : running %d : type %d : prog %d, video %d, audio %d, ttext %d : audio %s (up %d / seen %d)\n",
-			program->tsid, program->pnr, program->name, program->net, program->running,
-			program->type, program->p_pid, program->v_pid, program->a_pid, program->t_pid,
-			program->audio,
-			program->updated, program->seen
-	) ;
-}
-
 
 /* ----------------------------------------------------------------------- */
 static void print_streams(struct dvbmon *dvbmon)
@@ -236,21 +168,79 @@ struct psi_program *program ;
 struct list_head   *item;
 
 	fprintf(stderr, "\n==PROGRAMS==\n\n") ;
+    if (dvb_debug >= 25)
+    {
+    	fprintf(stderr, " &dvbmon->info->programs=>%p [next=%p, prev=%p]\n", &dvbmon->info->programs, dvbmon->info->programs.next, dvbmon->info->programs.prev) ;
+    }
     list_for_each(item,&dvbmon->info->programs)
     {
+	    if (dvb_debug >= 25)
+	    {
+	    	fprintf(stderr, "[item=%p next=%p, prev=%p] ", item, item->next, item->prev) ;
+	    }
         program = list_entry(item, struct psi_program, next);
         print_program(program) ;
     }
+	fprintf(stderr, "\n============\n\n") ;
 }
 
 
+
+
 /* ------------------------------------------------------------------------ */
 // CALLBACKS
 /* ------------------------------------------------------------------------ */
 
 /* ------------------------------------------------------------------------ */
 
-static int timeout = 20;
+/*
+4.4.2 Terrestrial delivery systems
+For terrestrial delivery systems bandwidth within a single transmitted TS is a valuable resource and in order to
+safeguard the bandwidth allocated to the primary services receivable from the actual multiplex, the following minimum
+repetition rates are specified in order to reflect the need to impose a limit on the amount of available bandwidth used for
+this purpose:
+a) all sections of the NIT shall be transmitted at least every 10 s;
+b) all sections of the BAT shall be transmitted at least every 10 s, if present;
+c) all sections of the SDT for the actual multiplex shall be transmitted at least every 2 s;
+d) all sections of the SDT for other TSs shall be transmitted at least every 10 s if present;
+e) all sections of the EIT Present/Following Table for the actual multiplex shall be transmitted at least every 2 s;
+f) all sections of the EIT Present/Following Tables for other TSs shall be transmitted at least every 20 s if
+present.
+
+The repetition rates for further EIT tables will depend greatly on the number of services and the quantity of related SI
+information. The following transmission intervals should be followed if practicable but they may be increased as the use
+of EIT tables is increased. The times are the consequence of a compromise between the acceptable provision of data to a
+viewer and the use of multiplex bandwidth.
+
+a) all sections of the EIT Schedule table for the first full day for the actual TS, shall be transmitted at least every
+10 s, if present;
+b) all sections of the EIT Schedule table for the first full day for other TSs, shall be transmitted at least every
+60 s, if present;
+c) all sections of the EIT Schedule table for the actual TS, shall be transmitted at least every 30 s, if present;
+d) all sections of the EIT Schedule table for other TSs, shall be transmitted at least every 300 s, if present;
+e) the TDT and TOT shall be transmitted at least every 30 s.
+
+*/
+
+/* ----------------------------------------------------------------------------- */
+
+//PAT								0x00 500 ms
+//TSDT								0x02 10 s [1]
+//NIT actual						0x10 10s / 25 ms
+//NIT other							0x10 10s / 25 ms
+//SDT actual						0x11 2s / 25 ms
+//SDT other							0x11 10s / 25 ms
+//BAT								0x11 10s / 25 ms
+//EIT actual present-following		0x12 2s / 25 ms [2]
+//EIT other present-following		0x12 10s / 25 ms [2]
+//TOT								0x14 30s / 25 ms
+//TDT								0x14 30s / 25 ms
+//PMT								ALL 500 ms
+
+static int scan_timeout_secs = 20 ;
+//static int nodata_timeout_secs = 3 ;	// should see *something* by then
+static int nodata_timeout_secs = 30 ;	// debugging!
+static int poll_timeout_ms = 1000 ;
 static int current;
 
 static void dvbwatch_tty(struct psi_info *info, int event,
@@ -260,31 +250,18 @@ static void dvbwatch_tty(struct psi_info *info, int event,
     struct psi_program *pr;
 
     // Uses static global:
-    // timeout
     // current
 
-
     switch (event)
     {
     case DVBMON_EVENT_SWITCH_TS:
 		if (dvbmon->verbose) fprintf(stderr,"  tsid  %5d\n",tsid);
 		current = tsid;
 		break;
+		
     case DVBMON_EVENT_UPDATE_PR:
-		pr = psi_program_get(info, tsid, pnr, 0);
-		if (!pr)
-			return;
-		if (tsid != current)
-			return;
-		if (pr->type != 1)
-			return;
-		if (0 == pr->v_pid)
-			return;
-		if (pr->name[0] == '\0')
-			return;
-
-		/* Hmm, get duplicates :-/ */
-		if (dvbmon->verbose) fprintf(stderr,"    pnr %5d  %s\n", pr->pnr, pr->name);
+		break ;
+		
     }
 }
 
@@ -298,23 +275,19 @@ static void dvbwatch_tty(struct psi_info *info, int event,
 /* ----------------------------------------------------------------------------- */
 static void table_open(struct dvbmon *dm, struct table *tab)
 {
-    if (tab->once && dm->tabfds >= dm->tablimit)
-	return;
+    if (tab->once && (dm->tabfds >= dm->tablimit))
+		return;
 
     tab->fd = dvb_demux_req_section(dm->dvb, -1, tab->pid, tab->sec, 0xff,
 				    tab->once, dm->timeout);
     if (-1 == tab->fd)
-	return;
+		return;
 
     dm->tabfds++;
 
-    //    tab->ch  = g_io_channel_unix_new(tab->fd);
-    //    tab->id  = g_io_add_watch(tab->ch, G_IO_IN, table_data, dm);
-//    tab->id = add_watch(tab->fd, POLLIN, table_data, dm) ;
-
     if (dm->tabdebug)
-	fprintf(stderr,"dvbmon: open:  %s %4d | fd=%d n=%d\n",
-		tab->name, tab->pid, tab->fd, dm->tabfds);
+		fprintf(stderr,"dvbmon: open:  %s %4d | fd=%d n=%d\n",
+			tab->name, tab->pid, tab->fd, dm->tabfds);
 }
 
 /* ----------------------------------------------------------------------------- */
@@ -324,9 +297,9 @@ static struct table* table_find(struct dvbmon *dm, int pid, int sec)
     struct list_head  *item;
 
     list_for_each(item,&dm->tables) {
-	tab = list_entry(item, struct table, next);
-	if (tab->pid == pid && tab->sec == sec)
-	    return tab;
+		tab = list_entry(item, struct table, next);
+		if (tab->pid == pid && tab->sec == sec)
+		    return tab;
     }
     return NULL;
 }
@@ -357,14 +330,14 @@ static void table_next(struct dvbmon *dm)
     struct list_head  *item;
 
     list_for_each(item,&dm->tables) {
-	tab = list_entry(item, struct table, next);
-	if (tab->fd != -1)
-	    continue;
-	if (tab->done)
-	    continue;
-	table_open(dm,tab);
-	if (dm->tabfds >= dm->tablimit)
-	    return;
+		tab = list_entry(item, struct table, next);
+		if (tab->fd != -1)
+		    continue;
+		if (tab->done)
+		    continue;
+		table_open(dm,tab);
+		if (dm->tabfds >= dm->tablimit)
+		    return;
     }
 }
 
@@ -387,8 +360,8 @@ static void table_add(struct dvbmon *dm, char *name, int pid, int sec,
     tab->done = 0;
     list_add_tail(&tab->next,&dm->tables);
     if (dm->tabdebug)
-	fprintf(stderr,"dvbmon: add:   %s %4d | sec=0x%02x once=%d\n",
-		tab->name, tab->pid, tab->sec, tab->once);
+	fprintf(stderr,"dvbmon: add:   %s %4d (0x%02x) | sec=0x%02x once=%d\n",
+		tab->name, tab->pid, tab->pid, tab->sec, tab->once);
 
     table_open(dm,tab);
 }
@@ -431,7 +404,6 @@ static int table_data_seen(struct dvbmon *dm, char *name, int id, int version)
 
     list_for_each(item,&dm->versions) {
 		ver = list_entry(item, struct version, next);
-	// Dies here because ver => 0x8! i.e. list is corrupt
 		if (ver->name == name && ver->id == id) {
 			if (ver->version == version)
 			seen = 1;
@@ -452,6 +424,54 @@ static int table_data_seen(struct dvbmon *dm, char *name, int id, int version)
 /* ----------------------------------------------------------------------------- */
 /* ----------------------------------------------------------------------------- */
 
+///* ----------------------------------------------------------------------------- */
+//void dvbmon_init_tables(struct dvbmon *dm, int o_nit, int o_sdt, int pmts)
+//{
+//    dm->tablimit = 3 + (o_nit ? 1 : 0) + (o_sdt ? 1 : 0) + pmts;
+//
+//if (dvb_debug >= 15) fprintf(stderr, "dvbmon_init_tables(onit=%d, osdt=%d, pmts=%d) : start\n", o_nit, o_sdt, pmts) ;
+//    
+//	table_add(dm, "pat",   0x00, 0x00, 0);
+//	table_add(dm, "nit",   0x10, 0x40, 0);
+//	table_add(dm, "sdt",   0x11, 0x42, 0);
+//	if (o_nit)
+//		table_add(dm, "nit",   0x10, 0x41, 0);
+//	if (o_sdt)
+//		table_add(dm, "sdt",   0x11, 0x46, 0);
+//
+//if (dvb_debug >= 15) fprintf(stderr, "dvbmon_init_tables() : done\n") ;
+//}
+//
+///* ----------------------------------------------------------------------------- */
+//void dvbmon_reset_tables(struct dvbmon *dm)
+//{
+//    struct list_head  *item, *safe;
+//    struct version    *ver;
+//    struct table      *tab;
+//
+//if (dvb_debug >= 15) fprintf(stderr, "dvbmon_reset_tables() : clear tables\n") ;
+//	// clear out tables
+//    list_for_each_safe(item,safe,&dm->tables) {
+//		tab = list_entry(item, struct table, next);
+//		table_del(dm, tab->pid, tab->sec);
+//    };
+//
+//if (dvb_debug >= 15) fprintf(stderr, "dvbmon_reset_tables() : clear versions\n") ;
+//    list_for_each_safe(item,safe,&dm->versions) {
+//		ver = list_entry(item, struct version, next);
+//		list_del(&ver->next);
+//		free(ver);
+//    };
+//
+//if (dvb_debug >= 15) fprintf(stderr, "dvbmon_reset_tables() : init tables\n") ;
+//	
+//	// restart
+//	dvbmon_init_tables(dm, /* other NIT */ 1,  /* other SDT */ 0, /* # PMTs */ 2) ;
+//
+//if (dvb_debug >= 15) fprintf(stderr, "dvbmon_reset_tables() : done\n") ;
+//	
+//}
+
 
 /* ----------------------------------------------------------------------------- */
 struct dvbmon*
@@ -472,8 +492,9 @@ dvbmon_init(struct dvb_state *dvb, int verbose, int o_nit, int o_sdt, int pmts)
     dm->dvb      = dvb;
     dm->info     = psi_info_alloc();
     if (dm->dvb) {
-		if (dm->verbose)
+		if (dm->verbose>1)
 			fprintf(stderr,"dvbmon: hwinit ok\n");
+
 		table_add(dm, "pat",   0x00, 0x00, 0);
 		table_add(dm, "nit",   0x10, 0x40, 0);
 		table_add(dm, "sdt",   0x11, 0x42, 0);
@@ -481,9 +502,15 @@ dvbmon_init(struct dvb_state *dvb, int verbose, int o_nit, int o_sdt, int pmts)
 			table_add(dm, "nit",   0x10, 0x41, 0);
 		if (o_sdt)
 			table_add(dm, "sdt",   0x11, 0x46, 0);
-		} else {
+
+
+	} else {
 		fprintf(stderr,"dvbmon: hwinit FAILED\n");
     }
+    
+    if (dvb_debug >= 15)
+		dm->tabdebug = 1 ;
+    
     return dm;
 }
 
@@ -578,7 +605,7 @@ void dvbmon_del_callback(struct dvbmon* dm, dvbmon_notify func, void *data)
 
 /* ------------------------------------------------------------------------ */
 /* ------------------------------------------------------------------------ */
-static int table_data(struct dvbmon *dm, struct table *tab, int verbose)
+static int table_data(struct dvbmon *dm, struct table *tab, int verbose, int tuned_freq)
 {
 struct list_head *item;
 struct psi_program *pr;
@@ -634,28 +661,28 @@ unsigned char buf[4096];
     switch (tab->sec) {
 		case 0x00: /* pat */
 			old_tsid = dm->info->tsid;
-			mpeg_parse_psi_pat(dm->info, buf, dm->verbose);
+			mpeg_parse_psi_pat(dm->info, buf, dm->verbose, tuned_freq);
 			if (old_tsid != dm->info->tsid)
 				call_callbacks(dm, DVBMON_EVENT_SWITCH_TS, dm->info->tsid, 0);
 			break;
 
 		case 0x02: /* pmt */
-			pr = psi_program_get(dm->info, dm->info->tsid, id, 0);
+			pr = psi_program_get(dm->info, dm->info->tsid, id, tuned_freq, 0);
 			if (!pr) {
 				if (dm->verbose) fprintf(stderr,"dvbmon: 404: tsid %d pid %d\n", dm->info->tsid, id);
 				break;
 			}
-			mpeg_parse_psi_pmt(pr, buf, dm->verbose);
+			mpeg_parse_psi_pmt(pr, buf, dm->verbose, tuned_freq);
 			break;
 
 		case 0x40: /* nit this  */
 		case 0x41: /* nit other */
-			mpeg_parse_psi_nit(dm->info, buf, dm->verbose);
+			mpeg_parse_psi_nit(dm->info, buf, dm->verbose, tuned_freq);
 			break;
 
 		case 0x42: /* sdt this  */
 		case 0x46: /* sdt other */
-			mpeg_parse_psi_sdt(dm->info, buf, dm->verbose);
+			mpeg_parse_psi_sdt(dm->info, buf, dm->verbose, tuned_freq);
 			break;
 
 		default:
@@ -676,7 +703,10 @@ unsigned char buf[4096];
 		list_for_each(item,&dm->info->programs) {
 			pr = list_entry(item, struct psi_program, next);
 			if (pr->seen && pr->p_pid)
+			{
 				table_add(dm, "pmt", pr->p_pid, 2, 1);
+				if (dvb_debug>=15) print_program(pr) ;
+			}
 			pr->seen = 0;
 		}
     }
@@ -695,8 +725,8 @@ unsigned char buf[4096];
 
 if (dvb_debug)
 {
-	fprintf(stderr, " + PROG: pnr %d tsid %d type %d name %s net %s : updated %d seen %d\n",
-			pr->pnr, pr->tsid, pr->type, pr->name, pr->net, pr->updated, pr->seen) ;
+	fprintf(stderr, " + PROG: pnr %d tsid %d type %d name %s net %s : updated %d seen %d [tuned=%d]\n",
+			pr->pnr, pr->tsid, pr->type, pr->name, pr->net, pr->updated, pr->seen, tuned_freq) ;
 }
 
 
@@ -715,113 +745,90 @@ if (dvb_debug)
 /* ------------------------------------------------------------------------ */
 /* ------------------------------------------------------------------------ */
 
-/* ------------------------------------------------------------------------ */
-static int dvb_tune_stream(struct dvb_state *dvb, struct psi_stream *stream, int timeout)
-{
-int rc;
-
-int frequency;
-int inversion=0;
-int bandwidth=0;
-int code_rate_high=0;
-int code_rate_low=0;
-int modulation=0;
-int transmission=0;
-int guard_interval=0;
-int hierarchy=0;
-
-//fprintf(stderr, "Tuning tsid %d (%s) - FREQ=%d\n", stream->tsid, stream->net, stream->frequency) ;
-//print_stream(stream) ;
-
-	// convert params
-	frequency = stream->frequency ;
-
-//	if (stream->polarization)
-//	{
-//		inversion = fe_vdr_bandwidth[ atoi(stream->polarization) ] ;
-//	}
-	if (stream->bandwidth)
-	{
-		bandwidth = atoi(stream->bandwidth) ;
-	}
-	if (stream->code_rate_hp)
-	{
-		code_rate_high = atoi(stream->code_rate_hp) ;
-	}
-	if (stream->code_rate_lp)
-	{
-		code_rate_low = atoi(stream->code_rate_lp) ;
-	}
-	if (stream->constellation)
-	{
-		modulation = atoi(stream->constellation) ;
-	}
-	if (stream->transmission)
-	{
-		transmission = atoi(stream->transmission) ;
-	}
-	if (stream->bandwidth)
-	{
-		guard_interval = atoi(stream->bandwidth) ;
-	}
-	if (stream->hierarchy)
-	{
-		hierarchy = atoi(stream->hierarchy) ;
-	}
-
-//fprintf(stderr, "TUNE: freq=%d inv=%d bw=%d rate_hi=%d rate_lo=%d mod=%d tx=%d guard=%d hier=%d\n",
-//		frequency,
-//		inversion,
-//		bandwidth,
-//		code_rate_high,
-//		code_rate_low,
-//		modulation,
-//		transmission,
-//		guard_interval,
-//		hierarchy
-//		) ;
-
-	// set tuning
-	rc = dvb_tune(dvb,
-			/* For frontend tuning */
-			frequency,
-			inversion,
-			bandwidth,
-			code_rate_high,
-			code_rate_low,
-			modulation,
-			transmission,
-			guard_interval,
-			hierarchy,
-			timeout) ;
-
-	return (rc) ;
-}
 
 /* ------------------------------------------------------------------------ */
+//clear freq list
+//
+//do
+//{
+//	tune to freq#	+ add to freq list
+//		get TSID + freqi	
+//		if 
+//			freqi == freq# -> tuned=1 ok=1
+//		else
+//			add to freq list
+//		end
+//		
+//		get other freqs
+//		add freqs to freq list
+//	
+//	get next freq
+//}
+//while (freqs left in list)
+//
 static void tty_scan(struct dvb_state *dvb, struct dvbmon *dvbmon)
 {
 time_t tuned;
-//char *sec;
 char *name;
 int num_freqs ;
 
 int ready ;
+int ready_total ;
 int i ;
 int rc ;
 
 int nfds=0 ; // TODO: just use dvbmon->tabfds
+
+
 struct pollfd *pollfds;
 struct table *tab = NULL;
 struct table **table_list = NULL;
 
-struct list_head   *item;
+struct list_head   *item, *safe;
 struct psi_stream *stream;
 
+struct freqitem *current_freqi, *stream_freqi ;
+
 // Uses static global:
-// timeout
+// poll_timeout_ms
+// scan_timeout_secs
 // current
 
+
+	// display settings
+	if (dvb_debug >= 2) dvb_frontend_tune_info(dvb);
+
+
+	// get current frequency info
+	current_freqi = freqitem_get(&dvb->p) ;
+
+	// mark the currently tuned in frequency as a success!
+	current_freqi->flags.tuned = 1 ;
+
+	// update signal quality
+	dvb_signal_quality(dvb,
+ 		&current_freqi->ber,
+		&current_freqi->snr,
+		&current_freqi->strength,
+		&current_freqi->uncorrected_blocks
+	) ;
+
+if (dvb_debug >= 1)
+{
+	fprintf_timestamp(stderr, "\n-- tty_scan() current FREQ %d (seen=%d, tuned=%d) -- \n",
+		dvb->p.frequency,
+		current_freqi->flags.seen,
+		current_freqi->flags.tuned
+	) ;
+}
+
+	// skip if already seen
+	if (current_freqi->flags.seen) return ;
+	
+	// mark the currently tuned as seen
+	current_freqi->flags.seen = 1 ;
+
+
     // Prepare for polling
     pollfds = (struct pollfd* )malloc(sizeof(struct pollfd) * dvbmon->tablimit);
     memset(pollfds, 0, sizeof(struct pollfd) * dvbmon->tablimit) ;
@@ -832,15 +839,52 @@ struct psi_stream *stream;
 	// start
     for (num_freqs=1;num_freqs;)
     {
+//    	// reset the scan tables
+//    	dvbmon_reset_tables(dvbmon) ;
+    	
+    	// reset global count (updated as we see data)
 		current = 0;
 
-		if (dvbmon->verbose>1) fprintf(stderr,"about to poll ..\n");
+		// get current frequency info
+		current_freqi = freqitem_get(&dvb->p) ;
 
-		/* fish data */
+		// mark the currently tuned in frequency as a success!
+		current_freqi->flags.tuned = 1 ;
+		current_freqi->flags.seen = 1 ;
+
+		// update signal quality
+		dvb_signal_quality(dvb,
+	 		&current_freqi->ber,
+			&current_freqi->snr,
+			&current_freqi->strength,
+			&current_freqi->uncorrected_blocks
+		) ;
+
+		
+		if (dvb_debug>1) fprintf_timestamp(stderr,"Current freqi = %d\n", current_freqi->frequency);
+
+		if (dvb_debug>10) fprintf(stderr,"about to poll ..\n");
+
+		// keep a count of data seen for this freq - allows us to abort early
+		ready_total = 0 ;
+
+		/* fish for data */
 		tuned = time(NULL);
-		while (time(NULL) - tuned < timeout)
+		while (time(NULL) - tuned < scan_timeout_secs)
 		{
-			if (dvbmon->verbose>1) fprintf(stderr,"Polling for data ...\n");
+			if (dvb_debug>10) fprintf(stderr,"Polling for data (timer=%d) ...\n", time(NULL) - tuned);
+
+//struct table {
+//    struct list_head    next;
+//    char                *name;
+//    int                 pid;
+//    int                 sec;
+//    int                 fd;
+//    int                 once;
+//    int                 done;
+//};
+
+if (dvb_debug >=14) fprintf(stderr, "Table List:\n") ;
 
 			// Get latest poll info
 			memset(pollfds, 0, sizeof(struct pollfd) * dvbmon->tablimit) ;
@@ -849,6 +893,8 @@ struct psi_stream *stream;
 			list_for_each(item, &dvbmon->tables)
 			{
 				tab = list_entry(item, struct table, next);
+if (dvb_debug >=14) fprintf(stderr, "(%d) Table %s : pid=0x%02x sec=0x%02x fd=%d once=%d done=%d\n", 
+									nfds, tab->name, tab->pid, tab->sec, tab->fd, tab->once, tab->done) ;
 				if (tab && (tab->fd>0) )
 				{
 					pollfds[nfds].fd = tab->fd ;
@@ -860,94 +906,208 @@ struct psi_stream *stream;
 					++nfds ;
 				}
 			}
+if (dvb_debug >=14) fprintf(stderr, "%d active entries\n\n", nfds) ;
 
 			// poll
-			ready = poll(pollfds, nfds, timeout);
+			ready = poll(pollfds, nfds, poll_timeout_ms);
+			ready_total += ready ;
 
-			if (dvbmon->verbose>1) fprintf(stderr," + ready=%d\n", ready);
+			if (dvb_debug>10) fprintf_timestamp(stderr," + ready=%d (total=%d)\n", ready, ready_total);
 
 			if (ready > 0)
 			{
 				// Check each fd
 				for(i=0; i<nfds; i++)
 				{
-					if (dvbmon->verbose>1) fprintf(stderr," %d : revents=%d\n", i, pollfds[i].revents);
+					if (dvb_debug>10) fprintf(stderr," %d : revents=0x%02x\n", i, pollfds[i].revents);
 
-					if (pollfds[i].revents == POLLIN)
+					if ( (pollfds[i].revents & POLLIN) == POLLIN)
 					{
-						if (dvbmon->verbose>2) fprintf(stderr," + dispatch..\n");
+						if (dvb_debug>10) fprintf(stderr," + dispatch..\n");
+if (dvb_debug >=14) fprintf(stderr, "Table[%d] %s : pid=0x%02x sec=0x%02x fd=%d once=%d done=%d\n", 
+	i, table_list[i]->name, table_list[i]->pid, table_list[i]->sec, table_list[i]->fd, table_list[i]->once, table_list[i]->done) ;
 
 						// fd is ready so dispatch
-						table_data(dvbmon, table_list[i], dvbmon->verbose) ;
+						table_data(dvbmon, table_list[i], dvbmon->verbose, current_freqi->frequency) ;
 					}
 				}
 			}
+			
+			if (ready == 0)
+			{
+				// have we seen anything on this freq?
+				if (ready_total == 0)
+				{
+					// abort early?
+					if (time(NULL) - tuned >= nodata_timeout_secs)
+					{
+if (dvb_debug >=14) fprintf_timestamp(stderr, "! Early abort...\n") ;
+						break ;
+					}	
+				}
+			}
+			
+			if (ready < 0)
+			{
+				perror("Poll fail") ;
+			}
 		}
 
-		if (dvbmon->verbose>1) fprintf(stderr,"done while loop ..\n");
+		if (dvb_debug>10) fprintf_timestamp(stderr,"done while loop ..\n");
+
 
 		if (!current)
 		{
-			if (dvbmon->verbose) fprintf(stderr,"Hmm, no data received. Frontend is%s locked.\n",
+			if (dvbmon->verbose) fprintf(stderr,"No data received. Frontend is%s locked.\n",
 										dvb_frontend_is_locked(dvb) ? "" : " not");
 		}
 
-//		fprintf(stderr, "\n\n----------------------------------------------------\nStreams so far:\n") ;
-//	    print_streams(dvbmon) ;
-//		fprintf(stderr, "\nStreams so far\n----------------------------------------------------\n\n") ;
+if (dvb_debug >= 10)
+{
+		fprintf(stderr, "\n\n----------------------------------------------------\nStreams so far:\n") ;
+	    print_streams(dvbmon) ;
+		fprintf(stderr, "----------------------------------------------------\n\n") ;
 
+		fprintf(stderr, "\n\n----------------------------------------------------\nFreq list so far:\n") ;
+	    print_freqs() ;
+		fprintf(stderr, "----------------------------------------------------\n\n") ;
+}
+
+if (dvb_debug >= 15)
+{
+		fprintf(stderr, "\n\n----------------------------------------------------\nPrograms so far:\n") ;
+	    print_programs(dvbmon) ;
+		fprintf(stderr, "----------------------------------------------------\n\n") ;
+}
+
+   		// Process streams
+		if (dvb_debug>=3) fprintf(stderr, "Processing streams...\n") ;
+	    list_for_each_safe(item,safe,&dvbmon->info->streams)
+	    {
+	    	stream = list_entry(item, struct psi_stream, next);
+
+		if (dvb_debug>=3) fprintf(stderr, "stream: freq %d -> tune=%d\n", stream->frequency, stream->tuned) ;
+
+			// only process un-tuned streams
+	    	if (!stream->tuned)
+	    	{
+	    		// check 'other frequency list'
+	    		if (stream->freq_list_len > 0)
+	    		{
+	    		int i ;
+	    		
+					if (dvb_debug>=3) 
+						fprintf(stderr, " + creating %d clones..\n", stream->freq_list_len) ;
+
+					for (i = 0; i < stream->freq_list_len; i++)
+					{
+					struct psi_stream *clone_stream;
+					struct freqitem *clone_stream_freqi ;
+						
+						// create a "cloned" stream with just a new frequency
+						clone_stream = psi_stream_newfreq(dvbmon->info, stream, stream->freq_list[i]) ;
+
+						// check to see if clone has been tuned to before
+			    		clone_stream_freqi = freqitem_get_from_stream(clone_stream) ;
+						if (dvb_debug>=3)
+						{
+							fprintf(stderr, " + Cloning tuned flag from: ") ;
+							print_freqi(clone_stream_freqi) ;
+						} 
+			    		clone_stream->tuned = clone_stream_freqi->flags.tuned ;
+
+						if (dvb_debug) 
+							fprintf(stderr, "clone stream: freq %d -> tune=%d\n", clone_stream->frequency, clone_stream->tuned) ;
+
+					}
+	    		}
+	    		
+	    		// check to see if this stream *has* actually been tuned already
+	    		stream_freqi = freqitem_get_from_stream(stream) ;
+				if (dvb_debug>=3)
+				{
+					fprintf(stderr, " + Copying tuned flag from: ") ;
+					print_freqi(stream_freqi) ;
+				} 
+	    		stream->tuned = stream_freqi->flags.tuned ;
+
+				if (dvb_debug) 
+					fprintf(stderr, "stream: freq %d -> tune=%d\n", stream->frequency, stream->tuned) ;
+	    	}
+	    }
+
+if (dvb_debug >= 10)
+{
+		fprintf(stderr, "\n\n----------------------------------------------------\nStreams after processing freq list:\n") ;
+	    print_streams(dvbmon) ;
+		fprintf(stderr, "----------------------------------------------------\n\n") ;
+
+		fprintf(stderr, "\n\n----------------------------------------------------\nFreq list so far:\n") ;
+	    print_freqs() ;
+		fprintf(stderr, "----------------------------------------------------\n\n") ;
+}
+
+	    
 		num_freqs=0;
 	    list_for_each(item,&dvbmon->info->streams)
 	    {
 	    	stream = list_entry(item, struct psi_stream, next);
 
+if (dvb_debug > 3) fprintf(stderr, "check stream  tsid %d freq %d Hz tuned=%d [num freqs=%d]\n", 
+						stream->tsid, stream->frequency, stream->tuned, num_freqs) ;
+
 	    	// Find next freq to select (if any left)
 	    	if (!stream->tuned)
 	    	{
-	    		// tune first
-	    		if (num_freqs==0)
-	    		{
-	    			stream->tuned=1 ;
-					if (dvbmon->verbose) fprintf(stderr, "Tuning tsid %d (%s)\n", stream->tsid, stream->net) ;
+				// map stream frequency onto freq list
+				stream_freqi = freqitem_get_from_stream(stream) ;
 
-					rc = dvb_tune_stream(dvb, stream, 100) ;
+if (dvb_debug > 3) fprintf(stderr, " + check stream freq seen=%d tuned=%d\n", stream_freqi->flags.seen, stream_freqi->flags.tuned) ;
+if (dvb_debug > 3) 
+{
+	fprintf(stderr, "Stream ") ;
+	print_freqi(stream_freqi) ;
+}
 
-					if (rc != 0 )
-					{
-						if (dvbmon->verbose) fprintf(stderr, " Tuning failed\n") ;
-						break ;
-					}
-	    		}
+//				if (!stream_freqi->flags.seen)
+//				{
+//if (dvb_debug>=10) fprintf(stderr, "stream: freq %d (%d) -> SKIP tune=%d\n", stream->frequency, stream_freqi->frequency, stream->tuned) ;
+//
+//	    		}
+//				else
+//				{
+//					if (dvbmon->verbose) fprintf(stderr, "Already seen freq %d - tsid %d (%s)\n", stream->frequency, stream->tsid, stream->net) ;
+//				}
 
-	    		// keep track of total number of freqs left to tune
-	    		++num_freqs ;
 	    	}
 	    }
 
 
-//fprintf(stderr, "# freqs left = %d\n", num_freqs) ;
+if (dvb_debug > 2) fprintf(stderr, "# freqs left = %d\n", num_freqs) ;
 
     }
 
+if (dvbmon->verbose >= 2)
+{
+    print_streams(dvbmon) ;
+    print_programs(dvbmon) ;
+}
+
     // Clean up
     free(table_list) ;
     free(pollfds) ;
 }
 
+
+
+
 /* ----------------------------------------------------------------------- */
+struct dvbmon *dvbmon ;
+
+
+// TODO: Pass in timeout value....
 struct dvbmon *dvb_scan_freqs(struct dvb_state *dvb, int verbose)
 {
-struct dvbmon *dvbmon ;
-struct psi_program *program ;
-struct psi_stream *stream;
-struct list_head   *item;
-
-
-	// Initialise the monitor
-    dvbmon = dvbmon_init(dvb, verbose, /* other NIT */ 1,  /* other SDT */ 1, /* # PMTs */ 2);
-
-	// set up scanning callback handler
-	dvbmon_add_callback(dvbmon,dvbwatch_tty, dvbmon);
 
 	// do scan
 	tty_scan(dvb, dvbmon);
@@ -956,3 +1116,24 @@ struct list_head   *item;
 	return dvbmon ;
 }
 
+
+/* ----------------------------------------------------------------------- */
+void dvb_scan_init(struct dvb_state *dvb, int verbose)
+{
+	// Initialise the monitor
+//    dvbmon = dvbmon_init(dvb, verbose, /* other NIT */ 1,  /* other SDT */ 1, /* # PMTs */ 2);
+    dvbmon = dvbmon_init(dvb, verbose, /* other NIT */ 1,  /* other SDT */ 0, /* # PMTs */ 2);
+//    dvbmon = dvbmon_init(dvb, verbose);
+//    dvbmon_reset_tables(dvbmon) ;
+
+	// set up scanning callback handler
+	dvbmon_add_callback(dvbmon,dvbwatch_tty, dvbmon);
+	
+}
+
+/* ----------------------------------------------------------------------- */
+void dvb_scan_end(struct dvb_state *dvb, int verbose)
+{
+	// clear out monitor
+	dvbmon_fini(dvbmon) ;
+}
